
from __future__ import absolute_import, print_function
import errno, glob, stat, subprocess
import os, sys
import pytest

from wvpytest import *

from bup import git, metadata
from bup import vfs
from bup.helpers import clear_errors, detect_fakeroot, is_superuser, resolve_parent
from bup.repo import LocalRepo
from bup.xstat import utime, lutime
import bup.helpers as helpers
from bup.compat import fsencode

lib_t_dir = os.path.dirname(fsencode(__file__))

top_dir = os.path.realpath(os.path.join(lib_t_dir, b'..', b'..'))

bup_path = top_dir + b'/bup'


def ex(*cmd):
    try:
        cmd_str = b' '.join(cmd)
        print(cmd_str, file=sys.stderr)
        rc = subprocess.call(cmd)
        if rc < 0:
            print('terminated by signal', - rc, file=sys.stderr)
            sys.exit(1)
        elif rc > 0:
            print('returned exit status', rc, file=sys.stderr)
            sys.exit(1)
    except OSError as e:
        print('subprocess call failed:', e, file=sys.stderr)
        sys.exit(1)


def setup_testfs():
    assert(sys.platform.startswith('linux'))
    # Set up testfs with user_xattr, etc.
    if subprocess.call([b'modprobe', b'loop']) != 0:
        return False
    subprocess.call([b'umount', b'testfs'])
    ex(b'dd', b'if=/dev/zero', b'of=testfs.img', b'bs=1M', b'count=32')
    ex(b'mke2fs', b'-F', b'-j', b'-m', b'0', b'testfs.img')
    ex(b'rm', b'-rf', b'testfs')
    os.mkdir(b'testfs')
    ex(b'mount', b'-o', b'loop,acl,user_xattr', b'testfs.img', b'testfs')
    # Hide, so that tests can't create risks.
    os.chown(b'testfs', 0, 0)
    os.chmod(b'testfs', 0o700)
    return True


def cleanup_testfs():
    subprocess.call([b'umount', b'testfs'])
    helpers.unlink(b'testfs.img')


def test_clean_up_archive_path():
    cleanup = metadata._clean_up_path_for_archive
    WVPASSEQ(cleanup(b'foo'), b'foo')
    WVPASSEQ(cleanup(b'/foo'), b'foo')
    WVPASSEQ(cleanup(b'///foo'), b'foo')
    WVPASSEQ(cleanup(b'/foo/bar'), b'foo/bar')
    WVPASSEQ(cleanup(b'foo/./bar'), b'foo/bar')
    WVPASSEQ(cleanup(b'/foo/./bar'), b'foo/bar')
    WVPASSEQ(cleanup(b'/foo/./bar/././baz'), b'foo/bar/baz')
    WVPASSEQ(cleanup(b'/foo/./bar///././baz'), b'foo/bar/baz')
    WVPASSEQ(cleanup(b'//./foo/./bar///././baz/.///'), b'foo/bar/baz/')
    WVPASSEQ(cleanup(b'./foo/./.bar'), b'foo/.bar')
    WVPASSEQ(cleanup(b'./foo/.'), b'foo')
    WVPASSEQ(cleanup(b'./foo/..'), b'.')
    WVPASSEQ(cleanup(b'//./..//.../..//.'), b'.')
    WVPASSEQ(cleanup(b'//./..//..././/.'), b'...')
    WVPASSEQ(cleanup(b'/////.'), b'.')
    WVPASSEQ(cleanup(b'/../'), b'.')
    WVPASSEQ(cleanup(b''), b'.')


def test_risky_path():
    risky = metadata._risky_path
    WVPASS(risky(b'/foo'))
    WVPASS(risky(b'///foo'))
    WVPASS(risky(b'/../foo'))
    WVPASS(risky(b'../foo'))
    WVPASS(risky(b'foo/..'))
    WVPASS(risky(b'foo/../'))
    WVPASS(risky(b'foo/../bar'))
    WVFAIL(risky(b'foo'))
    WVFAIL(risky(b'foo/'))
    WVFAIL(risky(b'foo///'))
    WVFAIL(risky(b'./foo'))
    WVFAIL(risky(b'foo/.'))
    WVFAIL(risky(b'./foo/.'))
    WVFAIL(risky(b'foo/bar'))
    WVFAIL(risky(b'foo/./bar'))


def test_clean_up_extract_path():
    cleanup = metadata._clean_up_extract_path
    WVPASSEQ(cleanup(b'/foo'), b'foo')
    WVPASSEQ(cleanup(b'///foo'), b'foo')
    WVFAIL(cleanup(b'/../foo'))
    WVFAIL(cleanup(b'../foo'))
    WVFAIL(cleanup(b'foo/..'))
    WVFAIL(cleanup(b'foo/../'))
    WVFAIL(cleanup(b'foo/../bar'))
    WVPASSEQ(cleanup(b'foo'), b'foo')
    WVPASSEQ(cleanup(b'foo/'), b'foo/')
    WVPASSEQ(cleanup(b'foo///'), b'foo///')
    WVPASSEQ(cleanup(b'./foo'), b'./foo')
    WVPASSEQ(cleanup(b'foo/.'), b'foo/.')
    WVPASSEQ(cleanup(b'./foo/.'), b'./foo/.')
    WVPASSEQ(cleanup(b'foo/bar'), b'foo/bar')
    WVPASSEQ(cleanup(b'foo/./bar'), b'foo/./bar')
    WVPASSEQ(cleanup(b'/'), b'.')
    WVPASSEQ(cleanup(b'./'), b'./')
    WVPASSEQ(cleanup(b'///foo/bar'), b'foo/bar')
    WVPASSEQ(cleanup(b'///foo/bar'), b'foo/bar')


def test_metadata_method(tmpdir):
    bup_dir = tmpdir + b'/bup'
    data_path = tmpdir + b'/foo'
    os.mkdir(data_path)
    ex(b'touch', data_path + b'/file')
    ex(b'ln', b'-s', b'file', data_path + b'/symlink')
    test_time1 = 13 * 1000000000
    test_time2 = 42 * 1000000000
    utime(data_path + b'/file', (0, test_time1))
    lutime(data_path + b'/symlink', (0, 0))
    utime(data_path, (0, test_time2))
    ex(bup_path, b'-d', bup_dir, b'init')
    ex(bup_path, b'-d', bup_dir, b'index', b'-v', data_path)
    ex(bup_path, b'-d', bup_dir, b'save', b'-tvvn', b'test', data_path)
    git.check_repo_or_die(bup_dir)
    with  LocalRepo() as repo:
        resolved = vfs.resolve(repo,
                               b'/test/latest' + resolve_parent(data_path),
                               follow=False)
        leaf_name, leaf_item = resolved[-1]
        m = leaf_item.meta
        WVPASS(m.mtime == test_time2)
        WVPASS(leaf_name == b'foo')
        contents = tuple(vfs.contents(repo, leaf_item))
        WVPASS(len(contents) == 3)
        WVPASSEQ(frozenset(name for name, item in contents),
                 frozenset((b'.', b'file', b'symlink')))
        for name, item in contents:
            if name == b'file':
                m = item.meta
                WVPASS(m.mtime == test_time1)
            elif name == b'symlink':
                m = item.meta
                WVPASSEQ(m.symlink_target, b'file')
                WVPASSEQ(m.size, 4)
                WVPASSEQ(m.mtime, 0)


def _first_err():
    if helpers.saved_errors:
        return str(helpers.saved_errors[0])
    return ''


def test_from_path_error(tmpdir):
    if is_superuser() or detect_fakeroot():
        return
    path = tmpdir + b'/foo'
    os.mkdir(path)
    m = metadata.from_path(path, archive_path=path, save_symlinks=True)
    WVPASSEQ(m.path, path)
    os.chmod(path, 0o000)
    metadata.from_path(path, archive_path=path, save_symlinks=True)
    if metadata.get_linux_file_attr:
        print('saved_errors:', helpers.saved_errors, file=sys.stderr)
        WVPASS(len(helpers.saved_errors) == 1)
        errmsg = _first_err()
        WVPASS(errmsg.startswith('read Linux attr'))
        clear_errors()


def _linux_attr_supported(path):
    # Expects path to denote a regular file or a directory.
    if not metadata.get_linux_file_attr:
        return False
    try:
        metadata.get_linux_file_attr(path)
    except OSError as e:
        if e.errno in (errno.ENOTTY, errno.ENOSYS, errno.EOPNOTSUPP):
            return False
        else:
            raise
    return True


def test_apply_to_path_restricted_access(tmpdir):
    if is_superuser() or detect_fakeroot():
        return
    if sys.platform.startswith('cygwin'):
        return # chmod 000 isn't effective.
    try:
        parent = tmpdir + b'/foo'
        path = parent + b'/bar'
        os.mkdir(parent)
        os.mkdir(path)
        clear_errors()
        if metadata.xattr:
            try:
                metadata.xattr.set(path, b'user.buptest', b'bup')
            except:
                print("failed to set test xattr")
                # ignore any failures here - maybe FS cannot do it
                pass
        m = metadata.from_path(path, archive_path=path, save_symlinks=True)
        WVPASSEQ(m.path, path)
        os.chmod(parent, 0o000)
        m.apply_to_path(path)
        print(b'saved_errors:', helpers.saved_errors, file=sys.stderr)
        expected_errors = ['utime: ']
        if m.linux_attr and _linux_attr_supported(tmpdir):
            expected_errors.append('Linux chattr: ')
        if metadata.xattr and m.linux_xattr:
            expected_errors.append("xattr.set ")
        WVPASS(len(helpers.saved_errors) == len(expected_errors))
        for i in range(len(expected_errors)):
            assert str(helpers.saved_errors[i]).startswith(expected_errors[i])
    finally:
        clear_errors()


def test_restore_over_existing_target(tmpdir):
    path = tmpdir + b'/foo'
    os.mkdir(path)
    dir_m = metadata.from_path(path, archive_path=path, save_symlinks=True)
    os.rmdir(path)
    open(path, 'w').close()
    file_m = metadata.from_path(path, archive_path=path, save_symlinks=True)
    # Restore dir over file.
    WVPASSEQ(dir_m.create_path(path, create_symlinks=True), None)
    WVPASS(stat.S_ISDIR(os.stat(path).st_mode))
    # Restore dir over dir.
    WVPASSEQ(dir_m.create_path(path, create_symlinks=True), None)
    WVPASS(stat.S_ISDIR(os.stat(path).st_mode))
    # Restore file over dir.
    WVPASSEQ(file_m.create_path(path, create_symlinks=True), None)
    WVPASS(stat.S_ISREG(os.stat(path).st_mode))
    # Restore file over file.
    WVPASSEQ(file_m.create_path(path, create_symlinks=True), None)
    WVPASS(stat.S_ISREG(os.stat(path).st_mode))
    # Restore file over non-empty dir.
    os.remove(path)
    os.mkdir(path)
    open(path + b'/bar', 'w').close()
    WVEXCEPT(Exception, file_m.create_path, path, create_symlinks=True)
    # Restore dir over non-empty dir.
    os.remove(path + b'/bar')
    os.mkdir(path + b'/bar')
    WVEXCEPT(Exception, dir_m.create_path, path, create_symlinks=True)


from bup.metadata import xattr
if xattr:
    def remove_selinux(attrs):
        return list(filter(lambda i: not i in (b'security.selinux', ),
                           attrs))

    def test_handling_of_incorrect_existing_linux_xattrs():
        if not is_superuser() or detect_fakeroot():
            pytest.skip('skipping test -- not superuser')
            return
        if not setup_testfs():
            pytest.skip('unable to load loop module; skipping dependent tests')
            return
        for f in glob.glob(b'testfs/*'):
            ex(b'rm', b'-rf', f)
        path = b'testfs/foo'
        open(path, 'w').close()
        xattr.set(path, b'foo', b'bar', namespace=xattr.NS_USER)
        m = metadata.from_path(path, archive_path=path, save_symlinks=True)
        xattr.set(path, b'baz', b'bax', namespace=xattr.NS_USER)
        m.apply_to_path(path, restore_numeric_ids=False)
        WVPASSEQ(remove_selinux(xattr.list(path)), [b'user.foo'])
        WVPASSEQ(xattr.get(path, b'user.foo'), b'bar')
        xattr.set(path, b'foo', b'baz', namespace=xattr.NS_USER)
        m.apply_to_path(path, restore_numeric_ids=False)
        WVPASSEQ(remove_selinux(xattr.list(path)), [b'user.foo'])
        WVPASSEQ(xattr.get(path, b'user.foo'), b'bar')
        xattr.remove(path, b'foo', namespace=xattr.NS_USER)
        m.apply_to_path(path, restore_numeric_ids=False)
        WVPASSEQ(remove_selinux(xattr.list(path)), [b'user.foo'])
        WVPASSEQ(xattr.get(path, b'user.foo'), b'bar')
        cleanup_testfs()




from __future__ import absolute_import, print_function

import bup_main, os, sys
if bup_main.env_pythonpath:
    if sys.version_info[0] < 3:
        os.environ['PYTHONPATH'] = bup_main.env_pythonpath
    else:
        os.environb[b'PYTHONPATH'] = bup_main.env_pythonpath
else:
    del os.environ['PYTHONPATH']

from importlib import import_module
from pkgutil import iter_modules
from subprocess import PIPE
from threading import Thread
import re, select, signal, subprocess

from bup import compat, path, helpers
from bup.compat import (
    environ,
    fsdecode,
    wrap_main
)
from bup.helpers import (
    columnate,
    handle_ctrl_c,
    log,
    tty_width
)
from bup.git import close_catpipes
from bup.io import byte_stream, path_msg
from bup.options import _tty_width
import bup.cmd

def maybe_import_early(argv):
    """Scan argv and import any modules specified by --import-py-module."""
    while argv:
        if argv[0] != '--import-py-module':
            argv = argv[1:]
            continue
        if len(argv) < 2:
            log("bup: --import-py-module must have an argument\n")
            exit(2)
        mod = argv[1]
        import_module(mod)
        argv = argv[2:]

maybe_import_early(compat.get_argv())

handle_ctrl_c()

cmdpath = path.cmddir()

# We manipulate the subcmds here as strings, but they must be ASCII
# compatible, since we're going to be looking for exactly
# b'bup-SUBCMD' to exec.

def usage(msg=""):
    log('Usage: bup [-?|--help] [-d BUP_DIR] [--debug] [--profile] '
        '<command> [options...]\n\n')
    common = dict(
        ftp = 'Browse backup sets using an ftp-like client',
        fsck = 'Check backup sets for damage and add redundancy information',
        fuse = 'Mount your backup sets as a filesystem',
        help = 'Print detailed help for the given command',
        index = 'Create or display the index of files to back up',
        on = 'Backup a remote machine to the local one',
        restore = 'Extract files from a backup set',
        save = 'Save files into a backup set (note: run "bup index" first)',
        tag = 'Tag commits for easier access',
        web = 'Launch a web server to examine backup sets',
    )

    log('Common commands:\n')
    for cmd,synopsis in sorted(common.items()):
        log('    %-10s %s\n' % (cmd, synopsis))
    log('\n')

    log('Other available commands:\n')
    cmds = set()
    for c in sorted(os.listdir(cmdpath)):
        if c.startswith(b'bup-') and c.find(b'.') < 0:
            cname = fsdecode(c[4:])
            if cname not in common:
                cmds.add(c[4:].decode(errors='backslashreplace'))
    # built-in commands take precedence
    for _, name, _ in iter_modules(path=bup.cmd.__path__):
        name = name.replace('_','-')
        if name not in common:
            cmds.add(name)

    log(columnate(sorted(cmds), '    '))
    log('\n')

    log("See 'bup help COMMAND' for more information on " +
        "a specific command.\n")
    if msg:
        log("\n%s\n" % msg)
    sys.exit(99)

def extract_argval(args):
    """Assume args (all elements bytes) starts with a -x, --x, or --x=,
argument that requires a value and return that value and the remaining
args.  Exit with an errror if the value is missing.

    """
    # Assumes that first arg is a valid arg
    arg = args[0]
    if b'=' in arg:
        val = arg.split(b'=')[1]
        if not val:
            usage('error: no value provided for %s option' % arg)
        return val, args[1:]
    if len(args) < 2:
        usage('error: no value provided for %s option' % arg)
    return args[1], args[2:]


args = compat.get_argvb()
if len(args) < 2:
    usage()

## Parse global options
help_requested = None
do_profile = False
bup_dir = None
args = args[1:]
subcmd = None
while args:
    arg = args[0]
    if arg in (b'-?', b'--help'):
        help_requested = True
        args = args[1:]
    elif arg in (b'-V', b'--version'):
        subcmd = [b'version']
        args = args[1:]
    elif arg in (b'-D', b'--debug'):
        helpers.buglvl += 1
        environ[b'BUP_DEBUG'] = b'%d' % helpers.buglvl
        args = args[1:]
    elif arg == b'--profile':
        do_profile = True
        args = args[1:]
    elif arg in (b'-d', b'--bup-dir') or arg.startswith(b'--bup-dir='):
        bup_dir, args = extract_argval(args)
    elif arg == b'--import-py-module' or arg.startswith(b'--import-py-module='):
        # Just need to skip it here
        _, args = extract_argval(args)
    elif arg.startswith(b'-'):
        usage('error: unexpected option "%s"'
              % arg.decode('ascii', 'backslashescape'))
    else:
        break

subcmd = subcmd or args

# Make BUP_DIR absolute, so we aren't affected by chdir (i.e. save -C, etc.).
if bup_dir:
    environ[b'BUP_DIR'] = os.path.abspath(bup_dir)

if len(subcmd) == 0:
    if help_requested:
        subcmd = [b'help']
    else:
        usage()

if help_requested and subcmd[0] != b'help':
    subcmd = [b'help'] + subcmd

if len(subcmd) > 1 and subcmd[1] == b'--help' and subcmd[0] != b'help':
    subcmd = [b'help', subcmd[0]] + subcmd[2:]

subcmd_name = subcmd[0]
if not subcmd_name:
    usage()

try:
    cmd_module_name = 'bup.cmd.' + subcmd_name.decode('ascii').replace('-', '_')
    cmd_module = import_module(cmd_module_name)
except ModuleNotFoundError as ex:
    if ex.name != cmd_module_name:
        raise ex
    cmd_module = None

if not cmd_module:
    subcmd[0] = os.path.join(cmdpath, b'bup-' + subcmd_name)
    if not os.path.exists(subcmd[0]):
        usage('error: unknown command "%s"' % path_msg(subcmd_name))

already_fixed = int(environ.get(b'BUP_FORCE_TTY', 0))
if subcmd_name in (b'mux', b'ftp', b'help', b'fuse'):
    fix_stdout = False
    fix_stderr = False
else:
    fix_stdout = not (already_fixed & 1) and os.isatty(1)
    fix_stderr = not (already_fixed & 2) and os.isatty(2)

if fix_stdout or fix_stderr:
    _ttymask = (fix_stdout and 1 or 0) + (fix_stderr and 2 or 0)
    environ[b'BUP_FORCE_TTY'] = b'%d' % _ttymask
    environ[b'BUP_TTY_WIDTH'] = b'%d' % _tty_width()


sep_rx = re.compile(br'([\r\n])')

def print_clean_line(dest, content, width, sep=None):
    """Write some or all of content, followed by sep, to the dest fd after
    padding the content with enough spaces to fill the current
    terminal width or truncating it to the terminal width if sep is a
    carriage return."""
    global sep_rx
    assert sep in (b'\r', b'\n', None)
    if not content:
        if sep:
            os.write(dest, sep)
        return
    for x in content:
        assert not sep_rx.match(x)
    content = b''.join(content)
    if sep == b'\r' and len(content) > width:
        content = content[:width]
    os.write(dest, content)
    if len(content) < width:
        os.write(dest, b' ' * (width - len(content)))
    if sep:
        os.write(dest, sep)

def filter_output(srcs, dests):
    """Transfer data from file descriptors in srcs to the corresponding
    file descriptors in dests print_clean_line until all of the srcs
    have closed.

    """
    global sep_rx
    assert all(isinstance(x, int) for x in srcs)
    assert len(srcs) == len(dests)
    srcs = tuple(srcs)
    dest_for = dict(zip(srcs, dests))
    pending = {}
    try:
        while srcs:
            ready_fds, _, _ = select.select(srcs, [], [])
            width = tty_width()
            for fd in ready_fds:
                buf = os.read(fd, 4096)
                dest = dest_for[fd]
                if not buf:
                    srcs = tuple([x for x in srcs if x is not fd])
                    print_clean_line(dest, pending.pop(fd, []), width)
                else:
                    split = sep_rx.split(buf)
                    while len(split) > 1:
                        content, sep = split[:2]
                        split = split[2:]
                        print_clean_line(dest,
                                         pending.pop(fd, []) + [content],
                                         width,
                                         sep)
                    assert len(split) == 1
                    if split[0]:
                        pending.setdefault(fd, []).extend(split)
    except BaseException as ex:
        pending_ex = ex
        # Try to finish each of the streams
        try:
            for fd, pending_items in pending.items():
                dest = dest_for[fd]
                width = tty_width()
                try:
                    print_clean_line(dest, pending_items, width)
                except (EnvironmentError, EOFError) as ex:
                    ex.__cause__ = pending_ex
                    pending_ex = ex
        finally:
            raise pending_ex

def import_and_run_main(module, args):
    if do_profile:
        import cProfile
        f = compile('module.main(args)', __file__, 'exec')
        cProfile.runctx(f, globals(), locals())
    else:
        module.main(args)


def run_module_cmd(module, args):
    if not (fix_stdout or fix_stderr):
        import_and_run_main(module, args)
        return
    # Interpose filter_output between all attempts to write to the
    # stdout/stderr and the real stdout/stderr (e.g. the fds that
    # connect directly to the terminal) via a thread that runs
    # filter_output in a pipeline.
    srcs = []
    dests = []
    real_out_fd = real_err_fd = stdout_pipe = stderr_pipe = None
    filter_thread = filter_thread_started = None
    try:
        if fix_stdout:
            sys.stdout.flush()
            stdout_pipe = os.pipe()  # monitored_by_filter, stdout_everyone_uses
            real_out_fd = os.dup(sys.stdout.fileno())
            os.dup2(stdout_pipe[1], sys.stdout.fileno())
            srcs.append(stdout_pipe[0])
            dests.append(real_out_fd)
        if fix_stderr:
            sys.stderr.flush()
            stderr_pipe = os.pipe()  # monitored_by_filter, stderr_everyone_uses
            real_err_fd = os.dup(sys.stderr.fileno())
            os.dup2(stderr_pipe[1], sys.stderr.fileno())
            srcs.append(stderr_pipe[0])
            dests.append(real_err_fd)

        filter_thread = Thread(name='output filter',
                               target=lambda : filter_output(srcs, dests))
        filter_thread.start()
        filter_thread_started = True
        import_and_run_main(module, args)
    finally:
        # Try to make sure that whatever else happens, we restore
        # stdout and stderr here, if that's possible, so that we don't
        # risk just losing some output.  Nest the finally blocks so we
        # try each one no matter what happens, and accumulate alll
        # exceptions in the pending exception __context__.
        try:
            try:
                try:
                    try:
                        real_out_fd is not None and \
                            os.dup2(real_out_fd, sys.stdout.fileno())
                    finally:
                        real_err_fd is not None and \
                            os.dup2(real_err_fd, sys.stderr.fileno())
                finally:
                    # Kick filter loose
                    stdout_pipe is not None and os.close(stdout_pipe[1])
            finally:
                stderr_pipe is not None and os.close(stderr_pipe[1])
        finally:
            close_catpipes()

    # There's no point in trying to join unless we finished the finally block.
    if filter_thread_started:
        filter_thread.join()


def run_subproc_cmd(args):

    c = (do_profile and [sys.executable, b'-m', b'cProfile'] or []) + args
    if not (fix_stdout or fix_stderr):
        os.execvp(c[0], c)

    sys.stdout.flush()
    sys.stderr.flush()
    out = byte_stream(sys.stdout)
    err = byte_stream(sys.stderr)
    p = None
    try:
        p = subprocess.Popen(c,
                             stdout=PIPE if fix_stdout else out,
                             stderr=PIPE if fix_stderr else err,
                             bufsize=4096, close_fds=True)
        # Assume p will receive these signals and quit, which will
        # then cause us to quit.
        for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT):
            signal.signal(sig, signal.SIG_IGN)

        srcs = []
        dests = []
        if fix_stdout:
            srcs.append(p.stdout.fileno())
            dests.append(out.fileno())
        if fix_stderr:
            srcs.append(p.stderr.fileno())
            dests.append(err.fileno())
        filter_output(srcs, dests)
        return p.wait()
    except BaseException as ex:
        if p and p.poll() == None:
            os.kill(p.pid, signal.SIGTERM)
            p.wait()
        raise


def run_subcmd(module, args):
    if module:
        run_module_cmd(module, args)
    else:
        run_subproc_cmd(args)

def main():
    wrap_main(lambda : run_subcmd(cmd_module, subcmd))

if __name__ == "__main__":
    main()




from __future__ import absolute_import, print_function
from collections import namedtuple
import mimetypes, os, posixpath, signal, stat, sys, time, webbrowser
from binascii import hexlify


from bup import options, git, vfs
from bup.helpers import (chunkyreader, debug1, format_filesize,
                         log, saved_errors)
from bup.path import resource_path
from bup.repo import LocalRepo
from bup.io import path_msg

try:
    from tornado import gen
    from tornado.httpserver import HTTPServer
    from tornado.ioloop import IOLoop
    from tornado.netutil import bind_unix_socket
    import tornado.web
except ImportError:
    log('error: cannot find the python "tornado" module; please install it\n')
    sys.exit(1)


# FIXME: right now the way hidden files are handled causes every
# directory to be traversed twice.


def http_date_from_utc_ns(utc_ns):
    return time.strftime('%a, %d %b %Y %H:%M:%S', time.gmtime(utc_ns / 10**9))


def _compute_breadcrumbs(path, show_hidden=False):
    """Returns a list of breadcrumb objects for a path."""
    breadcrumbs = []
    breadcrumbs.append((b'[root]', b'/'))
    path_parts = path.split(b'/')[1:-1]
    full_path = b'/'
    for part in path_parts:
        full_path += part + b"/"
        url_append = b""
        if show_hidden:
            url_append = b'?hidden=1'
        breadcrumbs.append((part, full_path+url_append))
    return breadcrumbs


def _contains_hidden_files(repo, dir_item):
    """Return true if the directory contains items with names other than
    '.' and '..' that begin with '.'

    """
    for name, item in vfs.contents(repo, dir_item, want_meta=False):
        if name in (b'.', b'..'):
            continue
        if name.startswith(b'.'):
            return True
    return False


def _dir_contents(repo, resolution, show_hidden=False):
    """Yield the display information for the contents of dir_item."""

    url_query = b'?hidden=1' if show_hidden else b''

    def display_info(name, item, resolved_item, display_name=None, omitsize=False):
        global opt
        # link should be based on fully resolved type to avoid extra
        # HTTP redirect.
        link = tornado.escape.url_escape(name, plus=False)
        if stat.S_ISDIR(vfs.item_mode(resolved_item)):
            link += '/'
        link = link.encode('ascii')

        if not omitsize:
            size = vfs.item_size(repo, item)
            if opt.human_readable:
                display_size = format_filesize(size)
            else:
                display_size = size
        else:
            display_size = None

        if not display_name:
            mode = vfs.item_mode(item)
            if stat.S_ISDIR(mode):
                display_name = name + b'/'
                display_size = None
            elif stat.S_ISLNK(mode):
                display_name = name + b'@'
                display_size = None
            else:
                display_name = name

        return display_name, link + url_query, display_size

    dir_item = resolution[-1][1]
    for name, item in vfs.contents(repo, dir_item):
        if not show_hidden:
            if (name not in (b'.', b'..')) and name.startswith(b'.'):
                continue
        if name == b'.':
            parent_item = resolution[-2][1] if len(resolution) > 1 else dir_item
            yield display_info(b'..', parent_item, parent_item, b'..', omitsize=True)
            continue
        res_item = vfs.ensure_item_has_metadata(repo, item, include_size=True)
        yield display_info(name, item, res_item)


class BupRequestHandler(tornado.web.RequestHandler):

    def initialize(self, repo=None):
        self.repo = repo

    def decode_argument(self, value, name=None):
        if name == 'path':
            return value
        return super().decode_argument(value, name)

    def get(self, path):
        return self._process_request(path)

    def head(self, path):
        return self._process_request(path)

    def _process_request(self, path):
        print('Handling request for %s' % path)
        sys.stdout.flush()
        # Set want_meta because dir metadata won't be fetched, and if
        # it's not a dir, then we're going to want the metadata.
        res = vfs.resolve(self.repo, path, want_meta=True)
        leaf_name, leaf_item = res[-1]
        if not leaf_item:
            self.send_error(404)
            return
        mode = vfs.item_mode(leaf_item)
        if stat.S_ISDIR(mode):
            self._list_directory(path, res)
        else:
            self._get_file(self.repo, path, res)

    def _list_directory(self, path, resolution):
        """Helper to produce a directory listing.

        Return value is either a file object, or None (indicating an
        error).  In either case, the headers are sent.
        """
        if not path.endswith(b'/') and len(path) > 0:
            print('Redirecting from %s to %s' % (path_msg(path), path_msg(path + b'/')))
            return self.redirect(path + b'/', permanent=True)

        hidden_arg = self.request.arguments.get('hidden', [0])[-1]
        try:
            show_hidden = int(hidden_arg)
        except ValueError as e:
            show_hidden = False

        self.render(
            'list-directory.html',
            path=path,
            breadcrumbs=_compute_breadcrumbs(path, show_hidden),
            files_hidden=_contains_hidden_files(self.repo, resolution[-1][1]),
            hidden_shown=show_hidden,
            dir_contents=_dir_contents(self.repo, resolution,
                                       show_hidden=show_hidden))
        return None

    @gen.coroutine
    def _get_file(self, repo, path, resolved):
        """Process a request on a file.

        Return value is either a file object, or None (indicating an error).
        In either case, the headers are sent.
        """
        file_item = resolved[-1][1]
        file_item = vfs.augment_item_meta(repo, file_item, include_size=True)
        meta = file_item.meta
        ctype = self._guess_type(path)
        self.set_header("Last-Modified", http_date_from_utc_ns(meta.mtime))
        self.set_header("Content-Type", ctype)

        self.set_header("Content-Length", str(meta.size))
        assert len(file_item.oid) == 20
        self.set_header("Etag", hexlify(file_item.oid))
        if self.request.method != 'HEAD':
            with vfs.fopen(self.repo, file_item) as f:
                it = chunkyreader(f)
                for blob in chunkyreader(f):
                    self.write(blob)
        raise gen.Return()

    def _guess_type(self, path):
        """Guess the type of a file.

        Argument is a PATH (a filename).

        Return value is a string of the form type/subtype,
        usable for a MIME Content-type header.

        The default implementation looks the file's extension
        up in the table self.extensions_map, using application/octet-stream
        as a default; however it would be permissible (if
        slow) to look inside the data to make a better guess.
        """
        base, ext = posixpath.splitext(path)
        if ext in self.extensions_map:
            return self.extensions_map[ext]
        ext = ext.lower()
        if ext in self.extensions_map:
            return self.extensions_map[ext]
        else:
            return self.extensions_map['']

    if not mimetypes.inited:
        mimetypes.init() # try to read system mime.types
    extensions_map = mimetypes.types_map.copy()
    extensions_map.update({
        '': 'text/plain', # Default
        '.py': 'text/plain',
        '.c': 'text/plain',
        '.h': 'text/plain',
        })


io_loop = None

def handle_sigterm(signum, frame):
    global io_loop
    debug1('\nbup-web: signal %d received\n' % signum)
    log('Shutdown requested\n')
    if not io_loop:
        sys.exit(0)
    io_loop.stop()


optspec = """
bup web [[hostname]:port]
bup web unix://path
--
human-readable    display human readable file sizes (i.e. 3.9K, 4.7M)
browser           show repository in default browser (incompatible with unix://)
"""

opt = None

def main(argv):
    global opt
    signal.signal(signal.SIGTERM, handle_sigterm)

    UnixAddress = namedtuple('UnixAddress', ['path'])
    InetAddress = namedtuple('InetAddress', ['host', 'port'])

    o = options.Options(optspec)
    opt, flags, extra = o.parse_bytes(argv[1:])

    if len(extra) > 1:
        o.fatal("at most one argument expected")

    if len(extra) == 0:
        address = InetAddress(host='127.0.0.1', port=8080)
    else:
        bind_url = extra[0]
        if bind_url.startswith('unix://'):
            address = UnixAddress(path=bind_url[len('unix://'):])
        else:
            addr_parts = extra[0].split(':', 1)
            if len(addr_parts) == 1:
                host = '127.0.0.1'
                port = addr_parts[0]
            else:
                host, port = addr_parts
            try:
                port = int(port)
            except (TypeError, ValueError) as ex:
                o.fatal('port must be an integer, not %r' % port)
            address = InetAddress(host=host, port=port)

    git.check_repo_or_die()

    settings = dict(
        debug = 1,
        template_path = resource_path(b'web').decode('utf-8'),
        static_path = resource_path(b'web/static').decode('utf-8'),
    )

    # Disable buffering on stdout, for debug messages
    try:
        sys.stdout._line_buffering = True
    except AttributeError:
        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)

    with LocalRepo() as repo:
        handlers = [ (r"(?P<path>/.*)", BupRequestHandler, dict(repo=repo))]
        application = tornado.web.Application(handlers, **settings)

        http_server = HTTPServer(application)
        io_loop_pending = IOLoop.instance()

        if isinstance(address, InetAddress):
            sockets = tornado.netutil.bind_sockets(address.port, address.host)
            http_server.add_sockets(sockets)
            print('Serving HTTP on %s:%d...' % sockets[0].getsockname()[0:2])
            if opt.browser:
                browser_addr = 'http://' + address[0] + ':' + str(address[1])
                io_loop_pending.add_callback(lambda : webbrowser.open(browser_addr))
        elif isinstance(address, UnixAddress):
            unix_socket = bind_unix_socket(address.path)
            http_server.add_socket(unix_socket)
            print('Serving HTTP on filesystem socket %r' % address.path)
        else:
            log('error: unexpected address %r', address)
            sys.exit(1)

        io_loop = io_loop_pending
        io_loop.start()

    if saved_errors:
        log('WARNING: %d errors encountered while saving.\n' % len(saved_errors))
        sys.exit(1)




from bup import git, options, vfs
from bup.compat import argv_bytes
from bup.gc import count_objects, find_live_objects
from bup.helpers import EXIT_FALSE, EXIT_TRUE, log, wrap_boolean_main
from bup.io import path_msg
from bup.repo import LocalRepo


optspec = """
bup validate-ref-links [ref...]
--
v,verbose       increase log output (can be used more than once)
"""

def validate(argv):
    o = options.Options(optspec)
    opt, flags, extra = o.parse_bytes(argv[1:])
    verbosity = opt.verbose

    git.check_repo_or_die()
    cat_pipe = git.cp()

    ref_missing = 0
    ref_info = []
    with LocalRepo() as repo:
        for ref in [argv_bytes(x) for x in extra]:
            # FIXME: unify with other commands and git: vfs:, etc.
            res = vfs.try_resolve(repo, ref, want_meta=False)
            # FIXME: if symlink, error(dangling)
            # FIXME: IOError ENOTDIR ELOOP
            _, leaf = res[-1]
            if not leaf:
                log(f'missing {path_msg(ref)}')
                ref_missing += 1
                continue
            kind = type(leaf)
            # FIXME: Root Tags FakeLink
            if kind in (vfs.Item, vfs.Chunky, vfs.RevList):
                ref_info.append((ref, leaf.oid))
            elif kind == vfs.Commit:
                ref_info.append((ref, leaf.coid))
            else:
                o.fatal(f"can't currently handle VFS {kind} for {path_msg(ref)}")

    found_missing = 0
    # Wanted all refs, or at least some specified weren't missing
    if not extra or (extra and ref_info):
        existing_count = count_objects(git.repo(b'objects/pack'), verbosity)
        if verbosity:
            log(f'found {existing_count} objects\n')

        if existing_count:
            with git.PackIdxList(git.repo(b'objects/pack')) as idxl:
                live_objects, live_trees, found_missing = \
                    find_live_objects(existing_count, cat_pipe, idxl, refs=ref_info,
                                      verbosity=verbosity, count_missing=True)
                live_objects.close()

    return EXIT_FALSE if (ref_missing + found_missing) else EXIT_TRUE

def main(argv):
    wrap_boolean_main(lambda: validate(argv))




import os, glob, sys

from bup import options, git, bloom
from bup.compat import argv_bytes, hexstr
from bup.helpers \
    import (add_error,
            debug1,
            log,
            note_error,
            progress,
            qprogress,
            saved_errors)
from bup.io import path_msg


optspec = """
bup bloom [options...]
--
ruin       ruin the specified bloom file (clearing the bitfield)
f,force    ignore existing bloom file and regenerate it from scratch
o,output=  output bloom filename (default: auto)
d,dir=     input directory to look for idx files (default: auto)
k,hashes=  number of hash functions to use (4 or 5) (default: auto)
c,check=   check given *.idx or *.midx file against the bloom filter
"""


def ruin_bloom(bloomfilename):
    rbloomfilename = git.repo_rel(bloomfilename)
    if not os.path.exists(bloomfilename):
        log(path_msg(bloomfilename) + '\n')
        add_error('bloom: %s not found to ruin\n' % path_msg(rbloomfilename))
        return
    with bloom.ShaBloom(bloomfilename, readwrite=True, expected=1) as b:
        b.map[16 : 16 + 2**b.bits] = b'\0' * 2**b.bits


def check_bloom(path, bloomfilename, idx):
    rbloomfilename = git.repo_rel(bloomfilename)
    ridx = git.repo_rel(idx)
    if not os.path.exists(bloomfilename):
        log('bloom: %s: does not exist.\n' % path_msg(rbloomfilename))
        return
    with bloom.ShaBloom(bloomfilename) as b:
        if not b.valid():
            add_error('bloom: %r is invalid.\n' % path_msg(rbloomfilename))
            return
        base = os.path.basename(idx)
        if base not in b.idxnames:
            log('bloom: %s does not contain the idx.\n' % path_msg(rbloomfilename))
            return
        if base == idx:
            idx = os.path.join(path, idx)
        log('bloom: bloom file: %s\n' % path_msg(rbloomfilename))
        log('bloom:   checking %s\n' % path_msg(ridx))
        oids = git.open_object_idx(idx)
        if not oids:
            note_error(f'bloom: ERROR: invalid index {path_msg(idx)}\n')
            return
        with oids:
            for oid in oids:
                if not b.exists(oid):
                    add_error('bloom: ERROR: object %s missing' % hexstr(oid))


_first = None
def do_bloom(path, outfilename, k, force):
    global _first
    assert k in (None, 4, 5)
    b = None
    try:
        if os.path.exists(outfilename) and not force:
            b = bloom.ShaBloom(outfilename)
            if not b.valid():
                debug1("bloom: Existing invalid bloom found, regenerating.\n")
                b.close()
                b = None

        add = []
        rest = []
        add_count = 0
        rest_count = 0
        for i, name in enumerate(glob.glob(b'%s/*.idx' % path)):
            progress('bloom: counting: %d\r' % i)
            with git.open_idx(name) as ix:
                ixbase = os.path.basename(name)
                if b is not None and (ixbase in b.idxnames):
                    rest.append(name)
                    rest_count += len(ix)
                else:
                    add.append(name)
                    add_count += len(ix)

        if not add:
            debug1("bloom: nothing to do.\n")
            return

        if b is not None:
            if len(b) != rest_count:
                debug1("bloom: size %d != idx total %d, regenerating\n"
                       % (len(b), rest_count))
                b, b_tmp = None, b
                b_tmp.close()
            elif k is not None and k != b.k:
                debug1("bloom: new k %d != existing k %d, regenerating\n"
                       % (k, b.k))
                b, b_tmp = None, b
                b_tmp.close()
            elif (b.bits < bloom.MAX_BLOOM_BITS[b.k] and
                  b.pfalse_positive(add_count) > bloom.MAX_PFALSE_POSITIVE):
                debug1("bloom: regenerating: adding %d entries gives "
                       "%.2f%% false positives.\n"
                       % (add_count, b.pfalse_positive(add_count)))
                b, b_tmp = None, b
                b_tmp.close()
            else:
                b, b_tmp = None, b
                b_tmp.close()
                b = bloom.ShaBloom(outfilename, readwrite=True,
                                   expected=add_count)
        if b is None: # Need all idxs to build from scratch
            add += rest
            add_count += rest_count
        del rest
        del rest_count

        msg = b is None and 'creating from' or 'adding'
        if not _first: _first = path
        dirprefix = (_first != path) and git.repo_rel(path) + b': ' or b''
        progress('bloom: %s%s %d file%s (%d object%s).\r'
            % (path_msg(dirprefix), msg,
               len(add), len(add)!=1 and 's' or '',
               add_count, add_count!=1 and 's' or ''))

        tfname = None
        if b is None:
            tfname = os.path.join(path, b'bup.tmp.bloom')
            b = bloom.create(tfname, expected=add_count, k=k)
        count = 0
        icount = 0
        for name in add:
            with git.open_idx(name) as ix:
                qprogress('bloom: writing %.2f%% (%d/%d objects)\r'
                          % (icount*100.0/add_count, icount, add_count))
                b.add_idx(ix)
                count += 1
                icount += len(ix)

    finally:  # This won't handle pending exceptions correctly in py2
        # Currently, there's an open file object for tfname inside b.
        # Make sure it's closed before rename.
        if b is not None: b.close()

    if tfname:
        os.rename(tfname, outfilename)


def main(argv):
    o = options.Options(optspec)
    opt, flags, extra = o.parse_bytes(argv[1:])

    if extra:
        o.fatal('no positional parameters expected')

    if not opt.check and opt.k and opt.k not in (4,5):
        o.fatal('only k values of 4 and 5 are supported')

    if opt.check:
        opt.check = argv_bytes(opt.check)

    git.check_repo_or_die()

    output = argv_bytes(opt.output) if opt.output else None
    path = opt.dir and argv_bytes(opt.dir) or git.repo(b'objects/pack')
    debug1('bloom: scanning %s\n' % path_msg(path))
    outfilename = output or os.path.join(path, b'bup.bloom')
    if opt.check:
        check_bloom(path, outfilename, opt.check)
    elif opt.ruin:
        ruin_bloom(outfilename)
    else:
        do_bloom(path, outfilename, opt.k, opt.force)

    if saved_errors:
        log('WARNING: %d errors encountered during bloom.\n' % len(saved_errors))
        sys.exit(1)
    elif opt.check:
        log('All tests passed.\n')




from __future__ import absolute_import, division, print_function
from binascii import hexlify
import sys, time

from bup import compat, hashsplit, git, options, client
from bup.compat import argv_bytes, environ, nullcontext
from bup.helpers import (add_error, hostname, log, parse_num,
                         qprogress, reprogress, saved_errors,
                         valid_save_name,
                         parse_date_or_fatal)
from bup.io import byte_stream
from bup.pwdgrp import userfullname, username


optspec = """
bup split [-t] [-c] [-n name] OPTIONS [--git-ids | filenames...]
bup split -b OPTIONS [--git-ids | filenames...]
bup split --copy OPTIONS [--git-ids | filenames...]
bup split --noop [-b|-t] OPTIONS [--git-ids | filenames...]
--
 Modes:
b,blobs    output a series of blob ids.  Implies --fanout=0.
t,tree     output a tree id
c,commit   output a commit id
n,name=    save the result under the given name
noop       split the input, but throw away the result
copy       split the input, copy it to stdout, don't save to repo
 Options:
r,remote=  remote repository path
d,date=    date for the commit (seconds since the epoch)
q,quiet    don't print progress messages
v,verbose  increase log output (can be used more than once)
git-ids    read a list of git object ids from stdin and split their contents
keep-boundaries  don't let one chunk span two input files
bench      print benchmark timings to stderr
max-pack-size=  maximum bytes in a single pack
max-pack-objects=  maximum number of objects in a single pack
fanout=    average number of blobs in a single tree
bwlimit=   maximum bytes/sec to transmit to server
#,compress=  set compression level to # (0-9, 9 is highest) [1]
"""


class NoOpPackWriter:
    def __init__(self):
        self.closed = False
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.close()
    def close(self):
        self.closed = True
    def __del__(self):
        assert self.closed
    def new_blob(self, content):
        return git.calc_hash(b'blob', content)
    def new_tree(self, shalist):
        return git.calc_hash(b'tree', git.tree_encode(shalist))

def opts_from_cmdline(argv):
    o = options.Options(optspec)
    opt, flags, extra = o.parse_bytes(argv[1:])
    opt.sources = extra

    if opt.name: opt.name = argv_bytes(opt.name)
    if opt.remote: opt.remote = argv_bytes(opt.remote)
    if opt.verbose is None: opt.verbose = 0

    if not (opt.blobs or opt.tree or opt.commit or opt.name or
            opt.noop or opt.copy):
        o.fatal("use one or more of -b, -t, -c, -n, --noop, --copy")
    if opt.copy and (opt.blobs or opt.tree):
        o.fatal('--copy is incompatible with -b, -t')
    if (opt.noop or opt.copy) and (opt.commit or opt.name):
        o.fatal('--noop and --copy are incompatible with -c, -n')
    if opt.blobs and (opt.tree or opt.commit or opt.name):
        o.fatal('-b is incompatible with -t, -c, -n')
    if extra and opt.git_ids:
        o.fatal("don't provide filenames when using --git-ids")
    if opt.verbose >= 2:
        git.verbose = opt.verbose - 1
        opt.bench = 1
    if opt.max_pack_size:
        opt.max_pack_size = parse_num(opt.max_pack_size)
    if opt.max_pack_objects:
        opt.max_pack_objects = parse_num(opt.max_pack_objects)
    if opt.fanout:
        opt.fanout = parse_num(opt.fanout)
    if opt.bwlimit:
        opt.bwlimit = parse_num(opt.bwlimit)
    if opt.date:
        opt.date = parse_date_or_fatal(opt.date, o.fatal)
    else:
        opt.date = time.time()

    opt.is_reverse = environ.get(b'BUP_SERVER_REVERSE')
    if opt.is_reverse:
        if opt.remote:
            o.fatal("don't use -r in reverse mode; it's automatic")
        if not opt.sources or opt.git_ids:
            o.fatal('"bup on ... split" does not support reading from standard input')

    if opt.name and not valid_save_name(opt.name):
        o.fatal("'%r' is not a valid branch name." % opt.name)

    return opt

def split(opt, files, parent, out, pack_writer):
    # Hack around lack of nonlocal vars in python 2
    total_bytes = [0]
    def prog(filenum, nbytes):
        total_bytes[0] += nbytes
        if filenum > 0:
            qprogress('Splitting: file #%d, %d kbytes\r'
                      % (filenum+1, total_bytes[0] // 1024))
        else:
            qprogress('Splitting: %d kbytes\r' % (total_bytes[0] // 1024))

    new_blob = pack_writer.new_blob
    new_tree = pack_writer.new_tree
    if opt.blobs:
        shalist = hashsplit.split_to_blobs(new_blob, files,
                                           keep_boundaries=opt.keep_boundaries,
                                           progress=prog)
        for sha, size, level in shalist:
            out.write(hexlify(sha) + b'\n')
            reprogress()
    elif opt.tree or opt.commit or opt.name:
        if opt.name: # insert dummy_name which may be used as a restore target
            mode, sha = \
                hashsplit.split_to_blob_or_tree(new_blob, new_tree, files,
                                                keep_boundaries=opt.keep_boundaries,
                                                progress=prog)
            splitfile_name = git.mangle_name(b'data', hashsplit.GIT_MODE_FILE, mode)
            shalist = [(mode, splitfile_name, sha)]
        else:
            shalist = \
                hashsplit.split_to_shalist(new_blob, new_tree, files,
                                           keep_boundaries=opt.keep_boundaries,
                                           progress=prog)
        tree = new_tree(shalist)
    else:
        last = 0
        it = hashsplit.hashsplit_iter(files,
                                      keep_boundaries=opt.keep_boundaries,
                                      progress=prog)
        for blob, level in it:
            hashsplit.total_split += len(blob)
            if opt.copy:
                sys.stdout.write(str(blob))
            megs = hashsplit.total_split // 1024 // 1024
            if not opt.quiet and last != megs:
                last = megs

    if opt.verbose:
        log('\n')
    if opt.tree:
        out.write(hexlify(tree) + b'\n')

    commit = None
    if opt.commit or opt.name:
        msg = b'bup split\n\nGenerated by command:\n%r\n' % compat.get_argvb()
        userline = b'%s <%s@%s>' % (userfullname(), username(), hostname())
        commit = pack_writer.new_commit(tree, parent, userline, opt.date,
                                        None, userline, opt.date, None, msg)
        if opt.commit:
            out.write(hexlify(commit) + b'\n')

    return commit

def main(argv):
    opt = opts_from_cmdline(argv)
    if opt.verbose >= 2:
        git.verbose = opt.verbose - 1
    if opt.fanout:
        hashsplit.fanout = opt.fanout
    if opt.blobs:
        hashsplit.fanout = 0
    if opt.bwlimit:
        client.bwlimit = opt.bwlimit

    start_time = time.time()

    sys.stdout.flush()
    out = byte_stream(sys.stdout)
    stdin = byte_stream(sys.stdin)

    writing = not (opt.noop or opt.copy)
    remote_dest = opt.remote or opt.is_reverse

    if writing or opt.git_ids:
        git.check_repo_or_die()

    if opt.git_ids:
        # the input is actually a series of git object ids that we should retrieve
        # and split.
        #
        # This is a bit messy, but basically it converts from a series of
        # CatPipe.get() iterators into a series of file-type objects.
        # It would be less ugly if either CatPipe.get() returned a file-like object
        # (not very efficient), or split_to_shalist() expected an iterator instead
        # of a file.
        cp = git.CatPipe()
        class IterToFile:
            def __init__(self, it):
                self.it = iter(it)
            def read(self, size):
                v = next(self.it, None)
                return v or b''
        def read_ids():
            while 1:
                line = stdin.readline()
                if not line:
                    break
                if line:
                    line = line.strip()
                try:
                    it = cp.get(line.strip())
                    next(it, None)  # skip the file info
                except KeyError as e:
                    add_error('error: %s' % e)
                    continue
                yield IterToFile(it)
        files = read_ids()
    else:
        # the input either comes from a series of files or from stdin.
        if opt.sources:
            files = (open(argv_bytes(fn), 'rb') for fn in opt.sources)
        else:
            files = [stdin]

    if remote_dest and writing:
        cli = repo = client.Client(opt.remote)
    else:
        cli = nullcontext()
        repo = git

    # cli creation must be last nontrivial command in each if clause above
    with cli:
        if opt.name and writing:
            refname = opt.name and b'refs/heads/%s' % opt.name
            oldref = repo.read_ref(refname)
        else:
            refname = oldref = None

        if not writing:
            pack_writer = NoOpPackWriter()
        elif not remote_dest:
            pack_writer = git.PackWriter(compression_level=opt.compress,
                                         max_pack_size=opt.max_pack_size,
                                         max_pack_objects=opt.max_pack_objects)
        else:
            pack_writer = cli.new_packwriter(compression_level=opt.compress,
                                             max_pack_size=opt.max_pack_size,
                                             max_pack_objects=opt.max_pack_objects)

        # packwriter creation must be last command in each if clause above
        with pack_writer:
            commit = split(opt, files, oldref, out, pack_writer)

        # pack_writer must be closed before we can update the ref
        if refname:
            repo.update_ref(refname, commit, oldref)

    secs = time.time() - start_time
    size = hashsplit.total_split
    if opt.bench:
        log('bup: %.2f kbytes in %.2f secs = %.2f kbytes/sec\n'
            % (size / 1024, secs, size / 1024 / secs))

    if saved_errors:
        log('WARNING: %d errors encountered while saving.\n' % len(saved_errors))
        sys.exit(1)




from __future__ import absolute_import
import os, time, random, subprocess, glob
import pytest

from bup import client, git, path
from bup.compat import bytes_from_uint, environ

def randbytes(sz):
    s = b''
    for i in range(sz):
        s += bytes_from_uint(random.randrange(0,256))
    return s


s1 = randbytes(10000)
s2 = randbytes(10000)
s3 = randbytes(10000)

IDX_PAT = b'/*.idx'


def test_server_split_with_indexes(tmpdir):
    environ[b'BUP_DIR'] = bupdir = tmpdir
    git.init_repo(bupdir)
    with git.PackWriter() as lw:
        lw.new_blob(s1)
    with client.Client(bupdir, create=True) as c, \
         c.new_packwriter() as rw:
        rw.new_blob(s2)
        rw.breakpoint()
        rw.new_blob(s1)


def test_multiple_suggestions(tmpdir):
    environ[b'BUP_DIR'] = bupdir = tmpdir
    git.init_repo(bupdir)

    with git.PackWriter() as lw:
        lw.new_blob(s1)
    with git.PackWriter() as lw:
        lw.new_blob(s2)
    assert len(glob.glob(git.repo(b'objects/pack'+IDX_PAT))) == 2

    with client.Client(bupdir, create=True) as c, \
         c.new_packwriter() as rw:

        assert len(glob.glob(c.cachedir+IDX_PAT)) == 0
        s1sha = rw.new_blob(s1)
        assert rw.exists(s1sha)
        s2sha = rw.new_blob(s2)

        # This is a little hacky, but ensures that we test the
        # code under test. First, flush to ensure that we've
        # actually sent all the command ('receive-objects-v2')
        # and their data to the server. This may be needed if
        # the output buffer size is bigger than the data (both
        # command and objects) we're writing. To see the need
        # for this, change the object sizes at the beginning
        # of this file to be very small (e.g. 10 instead of 10k)
        c.conn.outp.flush()

        # Then, check if we've already received the idx files.
        # This may happen if we're preempted just after writing
        # the data, then the server runs and suggests, and only
        # then we continue in PackWriter_Remote::_raw_write()
        # and check the has_input(), in that case we'll receive
        # the idx still in the rw.new_blob() calls above.
        #
        # In most cases though, that doesn't happen, and we'll
        # get past the has_input() check before the server has
        # a chance to respond - it has to actually hash the new
        # object here, so it takes some time. So also break out
        # of the loop if the server has sent something on the
        # connection.
        #
        # Finally, abort this after a little while (about one
        # second) just in case something's actually broken.
        n = 0
        while (len(glob.glob(c.cachedir+IDX_PAT)) < 2 and
               not c.conn.has_input() and n < 10):
            time.sleep(0.1)
            n += 1
        assert len(glob.glob(c.cachedir+IDX_PAT)) == 2 or c.conn.has_input()
        rw.new_blob(s2)
        assert rw.objcache.exists(s1sha)
        assert rw.objcache.exists(s2sha)
        rw.new_blob(s3)
        assert len(glob.glob(c.cachedir+IDX_PAT)) == 2
    assert len(glob.glob(c.cachedir+IDX_PAT)) == 3


def test_dumb_client_server(tmpdir):
    environ[b'BUP_DIR'] = bupdir = tmpdir
    git.init_repo(bupdir)
    open(git.repo(b'bup-dumb-server'), 'w').close()

    with git.PackWriter() as lw:
        lw.new_blob(s1)

    with client.Client(bupdir, create=True) as c, \
         c.new_packwriter() as rw:
        assert len(glob.glob(c.cachedir+IDX_PAT)) == 1
        rw.new_blob(s1)
        assert len(glob.glob(c.cachedir+IDX_PAT)) == 1
        rw.new_blob(s2)
    assert len(glob.glob(c.cachedir+IDX_PAT)) == 2


def test_midx_refreshing(tmpdir):
    environ[b'BUP_DIR'] = bupdir = tmpdir
    git.init_repo(bupdir)
    with client.Client(bupdir, create=True) as c, \
         c.new_packwriter() as rw:
        rw.new_blob(s1)
        p1base = rw.breakpoint()
        p1name = os.path.join(c.cachedir, p1base)
        s1sha = rw.new_blob(s1)  # should not be written; it's already in p1
        s2sha = rw.new_blob(s2)
        p2base = rw.close()
    p2name = os.path.join(c.cachedir, p2base)

    with git.PackIdxList(bupdir + b'/objects/pack') as pi:
        assert len(pi.packs) == 2
        pi.refresh()
        assert len(pi.packs) == 2
        assert sorted([os.path.basename(i.name) for i in pi.packs]) \
            == sorted([p1base, p2base])

        with git.open_idx(p1name) as p1, \
             git.open_idx(p2name) as p2:
            assert p1.exists(s1sha)
            assert not p2.exists(s1sha)
            assert p2.exists(s2sha)

        subprocess.call([path.exe(), b'midx', b'-f'])
        pi.refresh()
        assert len(pi.packs) == 1
        pi.refresh(skip_midx=True)
        assert len(pi.packs) == 2
        pi.refresh(skip_midx=False)
        assert len(pi.packs) == 1


def test_remote_parsing():
    tests = (
        (b':/bup', (b'file', None, None, b'/bup')),
        (b'file:///bup', (b'file', None, None, b'/bup')),
        (b'192.168.1.1:/bup', (b'ssh', b'192.168.1.1', None, b'/bup')),
        (b'ssh://192.168.1.1:2222/bup', (b'ssh', b'192.168.1.1', b'2222', b'/bup')),
        (b'ssh://[ff:fe::1]:2222/bup', (b'ssh', b'ff:fe::1', b'2222', b'/bup')),
        (b'bup://foo.com:1950', (b'bup', b'foo.com', b'1950', None)),
        (b'bup://foo.com:1950/bup', (b'bup', b'foo.com', b'1950', b'/bup')),
        (b'bup://[ff:fe::1]/bup', (b'bup', b'ff:fe::1', None, b'/bup')),)
    for remote, values in tests:
        assert client.parse_remote(remote) == values

    with pytest.raises(client.ClientError):
        client.parse_remote(b'http://asdf.com/bup')




from __future__ import absolute_import, print_function

from random import randint
from subprocess import CalledProcessError, check_output
from sys import stderr, stdout


from test.lib.wvpytest import wvpasseq

def rand_bytes(n):
    return bytes([randint(1, 255) for x in range(n)])

def test_argv():
    for trial in range(100):
        cmd = [b'dev/echo-argv-bytes', rand_bytes(randint(1, 32))]
        out = check_output(cmd)
        wvpasseq(b'\0\n'.join(cmd) + b'\0\n', out)




from __future__ import absolute_import
import io, math, os

from bup import _helpers, helpers
from bup._helpers import cat_bytes
from bup.compat import buffer
from bup.helpers import sc_page_size


_fmincore = getattr(helpers, 'fmincore', None)

BLOB_MAX = 8192*4   # 8192 is the "typical" blob size for bupsplit
BLOB_READ_SIZE = 8 * 1024 * 1024
MAX_PER_TREE = 256
progress_callback = None
fanout = 16

GIT_MODE_FILE = 0o100644
GIT_MODE_TREE = 0o40000
GIT_MODE_SYMLINK = 0o120000

# The purpose of this type of buffer is to avoid copying on peek(), get(),
# and eat().  We do copy the buffer contents on put(), but that should
# be ok if we always only put() large amounts of data at a time.
class Buf:
    def __init__(self):
        self.data = b''
        self.start = 0

    def put(self, s):
        if not self.data:
            self.data = s
            self.start = 0
        elif s:
            remaining = len(self.data) - self.start
            self.data = cat_bytes(self.data, self.start, remaining,
                                  s, 0, len(s))
            self.start = 0

    def peek(self, count):
        if count <= 256:
            return self.data[self.start : self.start + count]
        return buffer(self.data, self.start, count)

    def eat(self, count):
        self.start += count

    def get(self, count):
        if count <= 256:
            v = self.data[self.start : self.start + count]
        else:
            v = buffer(self.data, self.start, count)
        self.start += count
        return v

    def used(self):
        return len(self.data) - self.start


def _fadvise_pages_done(fd, first_page, count):
    assert(first_page >= 0)
    assert(count >= 0)
    if count > 0:
        _helpers.fadvise_done(fd,
                              first_page * sc_page_size,
                              count * sc_page_size)


def _nonresident_page_regions(status_bytes, incore_mask, max_region_len=None):
    """Return (start_page, count) pairs in ascending start_page order for
    each contiguous region of nonresident pages indicated by the
    mincore() status_bytes.  Limit the number of pages in each region
    to max_region_len."""
    assert(max_region_len is None or max_region_len > 0)
    start = None
    for i, x in enumerate(status_bytes):
        in_core = x & incore_mask
        if start is None:
            if not in_core:
                start = i
        else:
            count = i - start
            if in_core:
                yield (start, count)
                start = None
            elif max_region_len and count >= max_region_len:
                yield (start, count)
                start = i
    if start is not None:
        yield (start, len(status_bytes) - start)


def _uncache_ours_upto(fd, offset, first_region, remaining_regions):
    """Uncache the pages of fd indicated by first_region and
    remaining_regions that are before offset, where each region is a
    (start_page, count) pair.  The final region must have a start_page
    of None."""
    rstart, rlen = first_region
    while rstart is not None and (rstart + rlen) * sc_page_size <= offset:
        _fadvise_pages_done(fd, rstart, rlen)
        rstart, rlen = next(remaining_regions, (None, None))
    return (rstart, rlen)


def readfile_iter(files, progress=None):
    for filenum,f in enumerate(files):
        ofs = 0
        b = ''
        fd = rpr = rstart = rlen = None
        if _fmincore and hasattr(f, 'fileno'):
            try:
                fd = f.fileno()
            except io.UnsupportedOperation:
                pass
            if fd:
                mcore = _fmincore(fd)
                if mcore:
                    max_chunk = max(1, (8 * 1024 * 1024) / sc_page_size)
                    rpr = _nonresident_page_regions(mcore, helpers.MINCORE_INCORE,
                                                    max_chunk)
                    rstart, rlen = next(rpr, (None, None))
        while 1:
            if progress:
                progress(filenum, len(b))
            b = f.read(BLOB_READ_SIZE)
            ofs += len(b)
            if rpr:
                rstart, rlen = _uncache_ours_upto(fd, ofs, (rstart, rlen), rpr)
            if not b:
                break
            yield b
        if rpr:
            rstart, rlen = _uncache_ours_upto(fd, ofs, (rstart, rlen), rpr)


def _splitbuf(buf, basebits, fanbits):
    while 1:
        b = buf.peek(buf.used())
        (ofs, bits) = _helpers.splitbuf(b)
        if ofs:
            if ofs > BLOB_MAX:
                ofs = BLOB_MAX
                level = 0
            else:
                level = (bits-basebits)//fanbits  # integer division
            buf.eat(ofs)
            yield buffer(b, 0, ofs), level
        else:
            break
    while buf.used() >= BLOB_MAX:
        # limit max blob size
        yield buf.get(BLOB_MAX), 0


def _hashsplit_iter(files, progress):
    assert(BLOB_READ_SIZE > BLOB_MAX)
    basebits = _helpers.blobbits()
    fanbits = int(math.log(fanout or 128, 2))
    buf = Buf()
    for inblock in readfile_iter(files, progress):
        buf.put(inblock)
        for buf_and_level in _splitbuf(buf, basebits, fanbits):
            yield buf_and_level
    if buf.used():
        yield buf.get(buf.used()), 0


def _hashsplit_iter_keep_boundaries(files, progress):
    for real_filenum,f in enumerate(files):
        if progress:
            def prog(filenum, nbytes):
                # the inner _hashsplit_iter doesn't know the real file count,
                # so we'll replace it here.
                return progress(real_filenum, nbytes)
        else:
            prog = None
        for buf_and_level in _hashsplit_iter([f], progress=prog):
            yield buf_and_level


def hashsplit_iter(files, keep_boundaries, progress):
    if keep_boundaries:
        return _hashsplit_iter_keep_boundaries(files, progress)
    else:
        return _hashsplit_iter(files, progress)


total_split = 0
def split_to_blobs(makeblob, files, keep_boundaries, progress):
    global total_split
    for (blob, level) in hashsplit_iter(files, keep_boundaries, progress):
        sha = makeblob(blob)
        total_split += len(blob)
        if progress_callback:
            progress_callback(len(blob))
        yield (sha, len(blob), level)


def _make_shalist(l):
    ofs = 0
    l = list(l)
    total = sum(size for mode,sha,size, in l)
    vlen = len(b'%x' % total)
    shalist = []
    for (mode, sha, size) in l:
        shalist.append((mode, b'%0*x' % (vlen,ofs), sha))
        ofs += size
    assert(ofs == total)
    return (shalist, total)


def _squish(maketree, stacks, n):
    i = 0
    while i < n or len(stacks[i]) >= MAX_PER_TREE:
        while len(stacks) <= i+1:
            stacks.append([])
        if len(stacks[i]) == 1:
            stacks[i+1] += stacks[i]
        elif stacks[i]:
            (shalist, size) = _make_shalist(stacks[i])
            tree = maketree(shalist)
            stacks[i+1].append((GIT_MODE_TREE, tree, size))
        stacks[i] = []
        i += 1


def split_to_shalist(makeblob, maketree, files,
                     keep_boundaries, progress=None):
    sl = split_to_blobs(makeblob, files, keep_boundaries, progress)
    assert(fanout != 0)
    if not fanout:
        shal = []
        for (sha,size,level) in sl:
            shal.append((GIT_MODE_FILE, sha, size))
        return _make_shalist(shal)[0]
    else:
        stacks = [[]]
        for (sha,size,level) in sl:
            stacks[0].append((GIT_MODE_FILE, sha, size))
            _squish(maketree, stacks, level)
        #log('stacks: %r\n' % [len(i) for i in stacks])
        _squish(maketree, stacks, len(stacks)-1)
        #log('stacks: %r\n' % [len(i) for i in stacks])
        return _make_shalist(stacks[-1])[0]


def split_to_blob_or_tree(makeblob, maketree, files,
                          keep_boundaries, progress=None):
    shalist = list(split_to_shalist(makeblob, maketree,
                                    files, keep_boundaries, progress))
    if len(shalist) == 1:
        return (shalist[0][0], shalist[0][2])
    elif len(shalist) == 0:
        return (GIT_MODE_FILE, makeblob(b''))
    else:
        return (GIT_MODE_TREE, maketree(shalist))


def open_noatime(name):
    fd = _helpers.open_noatime(name)
    try:
        return os.fdopen(fd, 'rb', 1024*1024)
    except:
        try:
            os.close(fd)
        except:
            pass
        raise




from glob import glob
from os import environb, unlink
from subprocess import run
from sys import stderr

from bup import path

bup_exe = path.exe()

def runc(*args, **kwargs):
    assert 'check' not in kwargs
    run(*args, check='True', **kwargs)

def bupc(*args, **kwargs):
    cmd = [bup_exe] + list(args[0])
    print(cmd, file=stderr)
    run(cmd, *args[1:], check=True, **kwargs)

def test_missing_midx(tmpdir):
    bup_dir = tmpdir + b'/bup'
    pack_dir = bup_dir + b'/objects/pack'
    environb[b'GIT_DIR'] = bup_dir
    environb[b'BUP_DIR'] = bup_dir
    bupc(('init',))
    bupc(('index', 'test/sampledata/var/lib'))
    bupc(('save', '-n', 'save', 'test'))
    bupc(('index', 'test/sampledata/var/doc'))
    bupc(('save', '-n', 'save', 'test'))
    bupc(('midx', '-f'))
    midxs = glob(bup_dir + b'/objects/pack/*.midx')
    assert len(midxs) == 1
    midx = midxs[0]
    bupc(('midx', '--check', '-a'))
    idxs = glob(bup_dir + b'/objects/pack/*.idx')
    assert len(idxs) > 1
    unlink(idxs[0])
    bupc(('midx', '--check', '-a'))
